@@ -28,14 +28,16 @@ import boto3
 from cubicweb import Binary, set_log_methods
 from cubicweb.server.sources.storages import Storage
 from cubicweb.server.edition import EditedEntity
+from cubicweb.server.hook import DataOperationMixIn, Operation
 
 
 class S3Storage(Storage):
     is_source_callback = True
 
-    def __init__(self, bucket):
+    def __init__(self, bucket, suffix='.tmp'):
         self.s3cnx = self._s3_client()
         self.bucket = bucket
+        self.suffix = suffix
 
     @classmethod
     def _s3_client(cls):
@@ -48,32 +50,53 @@ class S3Storage(Storage):
     def callback(self, source, cnx, value):
         """see docstring for prototype, which vary according to is_source_callback
         """
-        key = source.binary_to_str(value).decode()
+        key = source.binary_to_str(value).decode('utf-8')
         try:
-            return Binary(self.s3cnx.get_object(
-                Bucket=self.bucket, Key=key)['Body'].read())
+            result = self.s3cnx.get_object(Bucket=self.bucket, Key=key)
+            self.info('Downloaded %s/%s from S3', self.bucket, key)
+            return Binary(result['Body'].read())
         except Exception as ex:
             source.critical("can't retrive S3 object %s: %s", value, ex)
             return None
 
     def entity_added(self, entity, attr):
         """an entity using this storage for attr has been added"""
+        if entity._cw.transaction_data.get('fs_importing'):
+            # fs_importing allows to change S3 key saved in database
+            entity._cw_dont_cache_attribute(attr, repo_side=True)
+            key = entity.cw_edited[attr].getvalue()
+            if PY3:
+                key = key.decode('utf-8')
+            try:
+                return Binary(self.s3cnx.get_object(
+                    Bucket=self.bucket, Key=key)['Body'].read())
+            except Exception:
+                return None
+
         binary = entity.cw_edited.pop(attr)
         if binary is not None:
             key = self.get_s3_key(entity, attr)
-            # bytes storage used to store S3's object key
-            binary_obj = Binary(key.encode())
-            entity.cw_edited.edited_attribute(attr, binary_obj)
-            # required workaround for boto3 bug
+            if key is None:
+                key = self.new_s3_key(entity, attr)
+            # save S3 key
+            entity.cw_edited.edited_attribute(
+                attr, Binary(key.encode('utf-8')))
+
+            # copy Binary value, workaround for boto3 bug
             # https://github.com/boto/s3transfer/issues/80
             # .read() is required since Binary can't wrap itself
+            binary.seek(0)
             buffer = Binary(binary.read())
-            self.debug('Upload object to S3')
-            # upload_fileobj should make automagically a multipart upload if
-            # needed
-            self.s3cnx.upload_fileobj(buffer, self.bucket, key)
+            binary.seek(0)
+            suffixed_key = self.suffixed_key(key)
+            self.s3cnx.upload_fileobj(buffer, self.bucket, suffixed_key)
             buffer.close()
-            self.info('Uploaded object %s.%s to S3', entity.eid, attr)
+
+            # move to final key in post commit event
+            S3AddFileOp.get_instance(entity._cw).add_data(
+                (self, key, entity.eid, attr))
+            self.info('Temporary uploaded %s.%s (%s/%s) to S3',
+                      entity.eid, attr, self.bucket, suffixed_key)
         return binary
 
     def entity_updated(self, entity, attr):
@@ -84,13 +107,14 @@ class S3Storage(Storage):
         """an entity using this storage for attr has been deleted"""
         if entity._cw.repo.config['s3-auto-delete']:
             key = self.get_s3_key(entity, attr)
-            self.info('Deleting object %s.%s (%s/%s) from S3',
+            if key is None:
+                # no key to remove
+                return
+            # delete key in a post commit event
+            S3DeleteFileOp.get_instance(entity._cw).add_data(
+                (self, key, entity.eid, attr))
+            self.info('Delaying deletion for %s.%s (%s/%s) in S3',
                       entity.eid, attr, self.bucket, key)
-            resp = self.s3cnx.delete_object(Bucket=self.bucket, Key=key)
-            if resp.get('ResponseMetadata', {}).get('HTTPStatusCode') >= 300:
-                self.error('S3 object deletion FAILED: %s', resp)
-            else:
-                self.debug('S3 object deletion OK: %s', resp)
 
     def migrate_entity(self, entity, attribute):
         """migrate an entity attribute to the storage"""
@@ -125,9 +149,9 @@ class S3Storage(Storage):
         if rset and rset.rows[0][0]:
             key = rset.rows[0][0].getvalue()
             if PY3:
-                key = key.decode()
+                key = key.decode('utf-8')
             return key
-        return self.new_s3_key(entity, attr)
+        return None
 
     def new_s3_key(self, entity, attr):
         """Generate a new key for given entity attr.
@@ -135,6 +159,48 @@ class S3Storage(Storage):
         This implemenation just return a random UUID"""
         return str(uuid.uuid1())
 
+    def suffixed_key(self, key):
+        return key + self.suffix
+
+
+class S3AddFileOp(DataOperationMixIn, Operation):
+    containercls = list
+
+    def postcommit_event(self):
+        for storage, key, eid, attr in self.get_data():
+            suffixed_key = storage.suffixed_key(key)
+            storage.s3cnx.copy_object(
+                Bucket=storage.bucket,
+                CopySource={'Bucket': storage.bucket, 'Key': suffixed_key},
+                Key=key)
+            storage.s3cnx.delete_object(
+                Bucket=storage.bucket, Key=suffixed_key)
+            self.info('Moved temporary object for %s.%s (%s/%s to %s/%s)'
+                      ' in S3', eid, attr, storage.bucket, suffixed_key,
+                      storage.bucket, key)
+
+    def rollback_event(self):
+        for storage, key, eid, attr in self.get_data():
+            suffixed_key = storage.suffixed_key(key)
+            storage.s3cnx.delete_object(
+                Bucket=storage.bucket, Key=suffixed_key)
+            self.info('Deleted temporary object for %s.%s (%s/%s) in S3',
+                      eid, attr, storage.bucket, suffixed_key)
+
+
+class S3DeleteFileOp(DataOperationMixIn, Operation):
+    containercls = list
+
+    def postcommit_event(self):
+        for storage, key, eid, attr in self.get_data():
+            self.info('Deleting object %s.%s (%s/%s) from S3',
+                      eid, attr, storage.bucket, key)
+            resp = storage.s3cnx.delete_object(Bucket=storage.bucket, Key=key)
+            if resp.get('ResponseMetadata', {}).get('HTTPStatusCode') >= 300:
+                self.error('S3 object deletion FAILED: %s', resp)
+            else:
+                self.debug('S3 object deletion OK: %s', resp)
+
 
 set_log_methods(S3Storage,
                 getLogger('cube.s3storage.storages.s3storage'))
